/* MCS barrier
 */

#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <sched.h>
#include <time.h>

#include "config.h"

typedef struct {
	uint8_t parentsense;
	uint8_t *parent;
	uint8_t *children[2];

	union {
		uint8_t single[4];
		uint32_t all;
	} havechild;

	union {
		uint8_t single[4];
		uint32_t all;
	} childnotready;

	uint8_t sense;
	uint8_t dummy;
	unsigned int id;
} barriernode;

typedef struct {
	unsigned int id;
	pthread_t thread;
	barriernode *n;
} threadinfo;

static void dobarrier (barriernode * const n) __attribute__((always_inline));

static void dobarrier (barriernode * const n) {
	/* spin */
	while (n->childnotready.all != 0x0) {
		sched_yield ();
	}
	n->childnotready.all = n->havechild.all;
	/* byte stores are atomic on amd64 */
	*n->parent = 0x0;
	if (n->id != 0) {
		while (n->parentsense != n->sense) {
			sched_yield ();
		}
	}
	*n->children[0] = n->sense;
	*n->children[1] = n->sense;
	n->sense = !n->sense;
}

void *worker (void *arg) {
	threadinfo *info = arg;
	const size_t runs = (info->id == 0) ? UNBALANCEDADDRUNS : ADDRUNS;

	for (size_t i = 0; i < BARRIERRUNS; i++) {
		volatile unsigned int foo = 0;
		for (size_t j = 0; j < runs; j++) {
			foo += j;
		}
		debug ("worker %u done, %u\n", info->id, foo);
		dobarrier (info->n);
	}
	return NULL;
}

int main () {
	threadinfo w[THREAD_NUM];
	barriernode nodes[THREAD_NUM];
	struct timespec start, end;

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		w[i].id = i;
		w[i].n = &nodes[i];

		/* init barrier */
		debug ("init for %u\n", i);
		for (size_t j = 0; j < 4; j++) {
			nodes[i].havechild.single[j] = (4*i+j < (THREAD_NUM-1)) ? true : false;
			debug ("havechild[%u]=%i\n", j, nodes[i].havechild.single[j]);
		}
		nodes[i].parent = (i != 0) ? &nodes[(i-1)/4].childnotready.single[(i-1)%4] : &nodes[i].dummy;
		if (i != 0) {
			debug ("parent = nodes[%u].childnotready[%u]\n", (i-1)/4, (i-1)%4);
		}
		nodes[i].children[0] = (2*i+1 < THREAD_NUM) ? &nodes[2*i+1].parentsense : &nodes[i].dummy;
		if (2*i+1 < THREAD_NUM) {
			debug ("children[0] = nodes[%u]\n", 2*i+1);
		}
		nodes[i].children[1] = (2*i+2 < THREAD_NUM) ? &nodes[2*i+2].parentsense : &nodes[i].dummy;
		if (2*i+2 < THREAD_NUM) {
			debug ("children[1] = nodes[%u]\n", 2*i+2);
		}
		nodes[i].childnotready.all = nodes[i].havechild.all;
		nodes[i].parentsense = 0x0;
		nodes[i].sense = 0x1;
		nodes[i].dummy = 0x0;
		nodes[i].id = i;
	}

	clock_gettime (CLOCK_MONOTONIC_RAW, &start);
	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_create (&w[i].thread, NULL, worker, &w[i]);
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_join (w[i].thread, NULL);
	}
	clock_gettime (CLOCK_MONOTONIC_RAW, &end);

	unsigned long long int t = (end.tv_sec - start.tv_sec) * 1000000000ULL +
			(end.tv_nsec - start.tv_nsec);
	printf ("%llu\n", t);
}

